"use strict";
/*
 * Module dependencies
 */
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var static_1 = require("../server/plugins/static");
var CopyWebpackPlugin = require("copy-webpack-plugin");
var rewrite_route_1 = require("../utils/rewrite-route");
var readPkg = require("read-pkg-up");
var webpack = require("webpack");
var glob_1 = require("../utils/glob");
var fs_1 = require("mz/fs");
var path_1 = require("path");
/**
 * Environment Variables
 */
function Environment() {
    return Object.keys(process.env).reduce(function (o, k) {
        o['process.env.' + k] = JSON.stringify(process.env[k]);
        return o;
    }, {});
}
// NOTE: these are required server-side
// so the server will require them directly
// and know where they are.
// (e.g. require('elmo/router'))
function ServerAliases(_projectRoot, _elmoRoot) {
    return {
        react: 'preact',
        'react-dom': 'preact',
        elmo: path_1.join(__dirname, '..', 'elmo.js'),
        style: path_1.join(__dirname, '..', 'lib', 'style', 'style.js')
    };
}
/**
 * Webpack Config
 */
function Webpack(input) {
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        var pkgPath, ELMO_ROOT, target, environment, action, routes, aliases, root, serverAliases, config, pages, publicPages, common, document, rules, plugins, constants;
        return tslib_1.__generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, readPkg({ cwd: __dirname })];
                case 1:
                    pkgPath = (_a.sent()).path;
                    ELMO_ROOT = path_1.dirname(pkgPath);
                    target = input.target, environment = input.environment, action = input.action, routes = input.routes, aliases = input.aliases;
                    root = path_1.resolve(process.cwd(), input.root || '.');
                    serverAliases = Object.assign(ServerAliases(root, ELMO_ROOT), aliases);
                    config = {
                        mode: environment,
                        name: 'ssr',
                        context: root,
                        module: { rules: [] },
                        plugins: [],
                        entry: {}
                    };
                    return [4 /*yield*/, glob_1.default(root, path_1.join('pages', '**'))
                        // filter out any page that contains an underscore in
                        // any part of their path
                    ];
                case 2:
                    pages = _a.sent();
                    publicPages = pages.filter(function (page) {
                        var rel = path_1.relative(path_1.join(root, 'pages'), page);
                        var parts = rel.split(path_1.sep);
                        for (var i = 0; i < parts.length; i++) {
                            var part = parts[i];
                            if (part[0] === '_')
                                return false;
                        }
                        return true;
                    });
                    if (publicPages.length === 0) {
                        return [2 /*return*/, null];
                    }
                    common = [];
                    // setup the entries
                    config.entry = publicPages.reduce(function (obj, page) {
                        var name = rewrite_route_1.default(path_1.join(root, 'pages'), '.html', page);
                        var bundle = [];
                        // add in the common bundles
                        bundle.push.apply(bundle, common);
                        obj[name] = bundle.concat(page);
                        return obj;
                    }, {});
                    document = pages.find(function (p) { return path_1.basename(p, path_1.extname(p)) === '_document'; });
                    config.entry['_document.html'] = document || path_1.join(__dirname, '../lib/_document.js');
                    config.target = 'node';
                    // this is to ensure that we use node's actual require(...)
                    // which is needed for styled-jsx, but should probably be the
                    // default anyway.
                    config.externals = [
                        function (_parent, request, fn) {
                            if (serverAliases[request]) {
                                return fn(null, serverAliases[request]);
                            }
                            else if (request[0] !== '.' && request[0] !== '/') {
                                // externalize all node_modules
                                return fn(null, request);
                            }
                            else {
                                // pass relative paths through (don't externalize)
                                return fn(null, undefined);
                            }
                        }
                    ];
                    // resolves
                    config.resolve = {
                        extensions: ['.ts', '.js', '.tsx', '.jsx'],
                        modules: [path_1.join(root, 'node_modules')]
                    };
                    // where to resolve loaders
                    config.resolveLoader = {
                        modules: [path_1.join(root, 'node_modules')]
                    };
                    /**
                     * Output
                     */
                    config.output = {
                        path: '/',
                        publicPath: '/',
                        // [name] ensures the path matches the entry key
                        filename: '[name]',
                        libraryTarget: 'umd'
                    };
                    rules = config.module.rules;
                    // support json files
                    rules.push({
                        test: /\.json$/,
                        use: ['json-loader']
                    });
                    // support loading .jsx files
                    rules.push({
                        test: /\.jsx/,
                        exclude: /(node_modules|bower_components)/,
                        loader: path_1.join(ELMO_ROOT, 'node_modules', 'babel-loader'),
                        options: {
                            plugins: [
                                // TODO: check that these plugins exist
                                path_1.join(ELMO_ROOT, 'node_modules', 'babel-plugin-transform-class-properties'),
                                [
                                    path_1.join(ELMO_ROOT, 'node_modules', 'babel-plugin-transform-react-jsx'),
                                    { pragma: 'h' }
                                ]
                            ]
                        }
                    });
                    // automatically add import { h } from 'preact' to jsx pages
                    // NOTE: we add this to tsx pages, but typescript will
                    // complain about not being able to resolve anyway, but
                    // it makes for simpler tests to include it
                    rules.push({
                        test: /\.[jt]sx/,
                        exclude: /(node_modules|bower_components)/,
                        loader: path_1.join(__dirname, '..', 'server', 'loaders', 'h.js')
                    });
                    // compile our project's typescript files
                    rules.push({
                        test: /\.tsx?$/,
                        exclude: /(node_modules|bower_components)/,
                        loader: path_1.join(__dirname, '..', 'server', 'loaders', 'typescript.js'),
                        query: { environment: environment, target: target, root: root }
                    });
                    /**
                     * Reverse plugins, webpack goes from bottom to top
                     * but top to bottom makes more sense in a config
                     */
                    rules = rules.reverse();
                    plugins = config.plugins || [];
                    constants = Environment();
                    // attach the environment to the NODE_ENV
                    constants['process.env.NODE_ENV'] = JSON.stringify(constants['process.env.NODE_ENV'] || environment);
                    // Define environment variables
                    plugins.push(new webpack.DefinePlugin(constants));
                    // Don't write files if there are errors
                    plugins.push(new webpack.NoEmitOnErrorsPlugin());
                    // use names instead of IDs
                    plugins.push(new webpack.NamedModulesPlugin());
                    // server-side
                    plugins.push(new static_1.default({
                        elmoRoot: ELMO_ROOT,
                        routes: routes,
                        environment: environment,
                        target: target,
                        root: root,
                        action: action
                    }));
                    return [4 /*yield*/, fs_1.exists(path_1.join(root, 'static'))];
                case 3:
                    // CopyWebpackPlugin will produce an error if this directory doesn't exist already
                    // so we check it first
                    if (_a.sent()) {
                        plugins.push(new CopyWebpackPlugin([
                            {
                                context: root,
                                from: './static',
                                to: './static',
                                ignore: ['.*']
                            }
                        ]));
                    }
                    return [4 /*yield*/, fs_1.exists(path_1.join(root, '_redirects'))];
                case 4:
                    // copy over the _redirects
                    if (_a.sent()) {
                        plugins.push(new CopyWebpackPlugin([
                            {
                                context: root,
                                from: '_redirects',
                                to: '_redirects',
                                toType: 'file'
                            }
                        ], { copyUnmodified: true }));
                    }
                    config.devtool = false;
                    return [2 /*return*/, config];
            }
        });
    });
}
exports.default = Webpack;
